/************************************************************
	FILE: lib_draw.c
          drawing and support for interactive
          manipulation of viewpoint
************************************************************/


#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>
#include <stdlib.h>
#include "tcl.h"
#include "glaux.h"
#include "gltk.h"

#include "lib_defs.h"
#include "lib_globals.h"
#include "lib_draw.h"
#include "lib_com.h"
#include "lib_math.h"
#include "lib_font.h"
#include "lib_user.h"
#include "lib_image.h"

static float Angle = 0.0;
static Matrix Vmat_orig;
static int Mouse[2];
static int Win[2];

int Mousemode = NO_MOTION;
int GL_winopen = 0;

/*********************************************************
    PROC: display()
    DOES: for running display in call-back mode
**********************************************************/

void display (void)
{
  extern void draw_world(void);
  extern void draw_world_texture(void);
  Matrix gl_vmat;

   glEnable(GL_DEPTH_TEST); 
   glClear((GLbitfield) (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT));
   glColor3f (1.0, 1.0, 1.0);
   mat_setup();

   user_draw();

   glLoadIdentity();
   glMatrixMode (GL_PROJECTION);
   glLoadIdentity();
   glDisable(GL_DEPTH_TEST);
   switch(Mode) {
     case TRANSLATE_MODE:
       draw_square();
       if (MAtprint)
	 mat_print();
       break;
     case ROTATE_MODE:
       draw_circle();
       if (MAtprint)
	 mat_print();
       break;
   }
   glFlush();
   if (DBuffer) 
     glXSwapBuffers(auxXDisplay(),auxXWindow());
}

/*********************************************************
    PROC: redisplay()
    DOES: allows tcl/tk commands to force redraw
**********************************************************/

int redisplay(ClientData clientData,
		   Tcl_Interp *interp,
		   int argc, char **argv) {
  switch(DRaw_mode) {
    case TEXTURE:
	  glEnable(GL_TEXTURE_2D);
	  break;
    case SOLID:
	  glDisable(GL_TEXTURE_2D);
	  break;
    case WIRE:
	  glDisable(GL_TEXTURE_2D);
	  break;
  }
  display();
  return TCL_OK;
}

/*********************************************************
    PROC: viewmat()
    DOES: loads or saves viewing matrix
**********************************************************/

int viewmat(ClientData clientData,
		   Tcl_Interp *interp,
		   int argc, char **argv)
{
  int n;

  if (argc!=3) {
	cerr(argc,argv,"Bad # args");
	return TCL_OK;
  }
  n = atoi(argv[2]) -1;
  if (n<0 || n>3) {
	cerr(argc,argv,"Number must be between 1 and 4");
	return TCL_OK;
  }
  switch( argv[1][0]) {
    case 'l':
	  tm_copy(Vmat,Vmat_mem[n]);
	  break;
	case 's':
	  tm_copy(Vmat_mem[n],Vmat);
	  save_vmat_mem();            /* save to file */
	  break;
  }
  return TCL_OK;
}

/*********************************************************
    PROC: save_vmat_mem()
    DOES: saves viewing matrices to memory
**********************************************************/

void save_vmat_mem()
{
  int n;
  STR fname = "views.mat";
  FILE *fp;
  int r,c;

  fp = fopen(fname,"w");
  if (!fp) {
	fprintf(stderr,"Unable to open file '%s' for write\n",fname);
	return;
  }
  for (n=0; n<4; n++) {
	for (r=0; r<4; r++) {
	  for (c=0; c<4; c++) {
		fprintf(fp,"%6.3f ",Vmat_mem[n][r][c]);
	  }
	  fprintf(fp,"\n");
	}
	fprintf(fp,"\n");
  }
  fclose(fp);
}

/*********************************************************
    PROC: load_vmat_mem()
    DOES: loads viewing matrices to memory
**********************************************************/

void load_vmat_mem()
{
  int n;
  STR fname = "views.mat";
  FILE *fp;
  int r,c;

  fp = fopen(fname,"r");
  if (!fp) {
	fprintf(stderr,"Unable to open file '%s' for read\n",fname);
	return;
  }
  for (n=0; n<4; n++) {
	for (r=0; r<4; r++) {
	  for (c=0; c<4; c++) {
		fscanf(fp,"%f ",&Vmat_mem[n][r][c]);
	  }
	}
  }
  fclose(fp);
}

/*********************************************************
    PROC: mat_print()
    DOES: prints ctm
**********************************************************/

void mat_print(void)
{
  Matrix tmat1, mat;
  STR row;
  int r;

  idenmat(tmat1);
  tmat1[2][3] = 5;
  tmmmult(mat,tmat1,Vmat);
  glColor3f(0.0, 0.0, 0.0);
  for (r=0; r<4; r++) {
    glRasterPos3f(0.0,0.9-r*0.1,0.0);
    sprintf(row,"%5.2f %5.2f %5.2f %5.2f\n",
	  mat[r][0], mat[r][1], mat[r][2], mat[r][3]);
    printString(row);
  }
/*  prtm(mat); */
}

/*********************************************************
    PROC:  draw_circle()
    DOES:  draws an approximation to a circle, indicating
           rotation mode
**********************************************************/

void draw_circle(void)
{
  int n;

  GLSTENCOLOUR;
  glBegin(GL_LINE_LOOP);
  for (n=0; n<CIRC_PTS; n++) {
    glVertex2fv(Circ[n]);
  }
  glEnd();
}

/*********************************************************
    PROC:  draw_square()
    DOES:  draws sights, to indicate translate mode
**********************************************************/

#define PC 0.9
#define PE 0.1

void draw_square(void)
{
  int n=0;
  static float v[12][2] = {
    -PC, -PC+PE,
    -PC,-PC,
    -PC+PE, -PC,
    PC-PE, -PC,
    PC, -PC,
    PC, -PC+PE,
    PC,PC-PE,
    PC,PC,
    PC-PE,PC,
    -PC+PE,PC,
    -PC,PC,
    -PC,PC-PE
  };

  GLSTENCOLOUR;
  glBegin(GL_LINE_STRIP);
  glVertex2fv(v[n++]);
  glVertex2fv(v[n++]);
  glVertex2fv(v[n++]);
  glEnd();
  glBegin(GL_LINE_STRIP);
  glVertex2fv(v[n++]);
  glVertex2fv(v[n++]);
  glVertex2fv(v[n++]);
  glEnd();
  glBegin(GL_LINE_STRIP);
  glVertex2fv(v[n++]);
  glVertex2fv(v[n++]);
  glVertex2fv(v[n++]);
  glEnd();
  glBegin(GL_LINE_STRIP);
  glVertex2fv(v[n++]);
  glVertex2fv(v[n++]);
  glVertex2fv(v[n++]);
  glEnd();
}

/*********************************************************
    PROC: myinit()
    DOES: 
**********************************************************/

void myinit (void) {
    glClearColor(1.0,1.0,1.0,0.0);
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);
    glShadeModel (GL_FLAT);
    makeRasterFont();
}

/*********************************************************
    PROC: tex_init()
    DOES: asks OpenGL to build MIP-maps, enables textures
**********************************************************/

void tex_init(void)
{
  if (TExtures) {
      glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
      glTexImage2D(GL_TEXTURE_2D, 0, 3, Image->h.xsize,
        Image->h.ysize, 0, GL_RGB, GL_UNSIGNED_BYTE,
        Image->data2);
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
      glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
      glEnable(GL_TEXTURE_2D);
  }
}

/*********************************************************
    PROC: myReshape
    DOES: Called when the window is first opened and whenever 
          the window is reconfigured (moved or resized).
**********************************************************/

void myReshape(int w, int h)
{
  Win[0] = w;
  Win[1] = h;
  glViewport (0, 0, w, h);	/*  define the viewport	*/
  mat_setup();
  display();
}

/*********************************************************
    PROC: mat_setup()
    DOES: sets up viewing & projection matrices
**********************************************************/

void mat_setup(void)
{
  Matrix gl_vmat;

    glMatrixMode (GL_PROJECTION);	/*  prepare for and then  */ 
    glLoadIdentity ();	/*  define the projection  */
    glFrustum (-1.0, 1.0, -1.0, 1.0, 1.5, 20.0);	/*  transformation  */
    glMatrixMode (GL_MODELVIEW);	/*  back to modelview matrix	*/
    tmtransp(gl_vmat, Vmat);
    glLoadMatrixf((GLfloat *) gl_vmat);
}

/*********************************************************
    PROC: OpenGL_setup()
    DOES: sets up window for OpenGL
**********************************************************/

int OpenGL_setup(void)
{
  if (DBuffer) {
    auxInitDisplayMode ( (GLenum) (AUX_DOUBLE | AUX_RGB | AUX_DEPTH));
  } else {
    auxInitDisplayMode ( (GLenum) (AUX_RGB | AUX_DEPTH));
  }
  auxInitPosition (0, 0, 500, 500);
  auxInitWindow ("GLS");
  myinit ();
  auxReshapeFunc (myReshape);
  auxMouseFunc(AUX_LEFTBUTTON, AUX_MOUSEDOWN, mouse_action);
  auxMouseFunc(AUX_MIDDLEBUTTON, AUX_MOUSEDOWN, mouse2_action);
  auxMouseFunc(AUX_LEFTBUTTON, AUX_MOUSEUP, mouse_up);
  auxMouseFunc(AUX_MIDDLEBUTTON, AUX_MOUSEUP, mouse_up);

  user_gl_init();      // initialize any user-defined callbacks

  GL_winopen = 1;
  tkDisplayFunc(display);
}

/*********************************************************
    PROC: toggle_view()
    DOES: toggles between rotate & translate modes
**********************************************************/

int toggle_view(ClientData clientData,
		   Tcl_Interp *interp,
		   int argc, char **argv)
{
  switch(argv[1][0]) {
    case 'r':
      Mode = ROTATE_MODE;
      break;
    case 't':
      Mode = TRANSLATE_MODE;
      break;
	}
  display();
  return TCL_OK;
}

/*********************************************************
    PROC: check_mouse()
    DOES: if mouse is down, updates for translation or rotation
**********************************************************/

void check_mouse(int x, int y)
{
  tmcopy(Vmat,Vmat_orig);    /* restore matrix */
  switch(Mousemode) {
    case NO_MOTION:
	  return;
    case ROTATE_MOTION:
	  rotate(x,y);
	  break;
    case TRANS_MOTION:
	  Vmat[0][3] += (x - Mouse[0])*0.008;
	  Vmat[1][3] -= (y - Mouse[1])*0.008;
	  break;
    case SCALE_MOTION:
	  scale_motion(x,y);
	  break;
    case Z_MOTION:
      Vmat[2][3] += (Mouse[1] - y)*0.088;
	  break;
  }
  display();
  glFinish();

}

/*********************************************************
    PROC: mouse_up(event)
    DOES: takes action for mouse up
**********************************************************/

void mouse_up(AUX_EVENTREC *event)
{
  int x,y;

  x = event->data[AUX_MOUSEX];   /* grab (x,y) from event queue */
  y = event->data[AUX_MOUSEY];
  check_mouse(x,y);              /* do update with the last x,y */
  Mousemode = NO_MOTION;         /* done tracking mouse for now */
}

/*********************************************************
    PROC: mouse_action(event)
    DOES: take translate or rotate action
**********************************************************/

void mouse_action(AUX_EVENTREC *event)
{
  Mouse[0] = event->data[AUX_MOUSEX];
  Mouse[1] = event->data[AUX_MOUSEY];
  tmcopy(Vmat_orig, Vmat);             /* save copy of original */
  tkDisplayFunc(NULL);
  switch(Mode) {
    case ROTATE_MODE:
	  Mousemode = ROTATE_MOTION;
      break;
    case TRANSLATE_MODE:
	  Mousemode = TRANS_MOTION;
      break;
  }
  tkIdleFunc(NULL);
  tkDisplayFunc(display);
}

/*********************************************************
    PROC: mouse2_action(event)
    DOES: take a scaling (rotate) or panning (translate) action
**********************************************************/

void mouse2_action(AUX_EVENTREC *event)
{
  Mouse[0] = event->data[AUX_MOUSEX];
  Mouse[1] = event->data[AUX_MOUSEY];
  tmcopy(Vmat_orig, Vmat);             /* save copy of original */
  tkDisplayFunc(NULL);
  switch(Mode) {
    case ROTATE_MODE:
	  Mousemode = SCALE_MOTION;
      break;
    case TRANSLATE_MODE:
	  Mousemode = Z_MOTION;
      break;
  }
  tkIdleFunc(NULL);
  tkDisplayFunc(display);
}

/*********************************************************
    PROC: rotate()
    DOES: rotates by the given amount
**********************************************************/

void rotate(int x, int y)
{
  Matrix tmat1, tmat2;
  Matrix mat1,mat2;
  FVEC quat;
  Matrix qm;

  calc_quat(quat, Mouse[0], Mouse[1], x, y);
  rot_mat2(quat,qm);
  idenmat(tmat1);
  idenmat(tmat2);
  tmat1[2][3] = 5;
  tmat2[2][3] = -5;
  tmmmult(mat1,qm,tmat1);
  tmmmult(mat2,tmat2,mat1);
  tmmmult(mat1,mat2,Vmat);
  tmcopy(Vmat,mat1);
}

/*********************************************************
    PROC: scale_motion(int x, int y)
    DOES: scaling motion with middle mouse when in rotate mode
**********************************************************/

void scale_motion(int x, int y)
{
  Matrix mat, mat1, mat2, tmat1, tmat2;

  idenmat(mat);
  mat[0][0] = mat[1][1] = mat[2][2] = 1.0 + 0.008*(Mouse[1] - y);

  idenmat(tmat1);
  idenmat(tmat2);
  tmat1[2][3] = 5;
  tmat2[2][3] = -5;
  tmmmult(mat1,mat,tmat1);
  tmmmult(mat2,tmat2,mat1);
  tmmmult(mat1,mat2,Vmat);
  tmcopy(Vmat,mat1);
}

/************************************************************
	PROC:	calc_p()
	DOES:	determines the point on the unit sphere
************************************************************/

void calc_p(int mousex, int mousey, float p[])
{
	float r,s;

	p[0] = 2.0*(mousex - 0.5*Win[0])/Win[0];
	p[1] = -2.0*(mousey - 0.5*Win[1])/Win[1];;
	r = p[0]*p[0] + p[1]*p[1];
	if (r>1.0) {
		s = 1.0/sqrt(r);
		p[0] *= s;
		p[1] *= s;
		p[2] = 0.0;
	} else {
		p[2] = sqrt(1.0 - r);
	}
}

/************************************************************
	PROC:	calc_quat()
	DOES:	calculates quaternion given mouse coords
	FROM:	Graphics Interface '92, "ARCBALL: ...",
			page 154. Note: there is a bug on this page;
			It should be q.w = -V3_DOT[p0,p1], as per
			ACM SIG '89 Tutorial Vol 23, p. 178 and also p. 191
************************************************************/

void calc_quat(float q[],
			   int mousex0, int mousey0,
			   int mousex1, int mousey1)
{
	int n;
	FVEC p0, p1;

	calc_p(mousex0,mousey0,p0);
	calc_p(mousex1,mousey1,p1);
	q[1] = p0[1]*p1[2] - p1[1]*p0[2];
	q[2] = p1[0]*p0[2] - p0[0]*p1[2];
	q[3] = p0[0]*p1[1] - p1[0]*p0[1];
	q[0] = (p0[0]*p1[0] + p0[1]*p1[1] + p0[2]*p1[2]);
	if (q[0]<-1.0) q[0]= -1.0;
}

