/************************************************************
	FILE: lib_math.c

************************************************************/

#include <stdlib.h>
#include "lib_defs.h"
#include "lib_globals.h"
#include "lib_math.h"

/*********************************************************
    PROC:  plotstate(fp,x,n)
    DOES:  plots the current state 
**********************************************************/

plotstate(FILE *fp, float *x, int n)
{
  int d;

/*
  fp = stdout;
*/
  if (n>20) {
    fprintf(fp,"shwn 21 ");
    for (d=20; d<n; d++) {
      fprintf(fp," %6.1f",x[d]);
    }
    fprintf(fp,"\n");
    fprintf(fp,"show ");
    for (d=0; d<20; d++) {
      fprintf(fp," %6.1f",x[d]);
    }
    fprintf(fp,"\n");
  } else {
    fprintf(fp,"show ");
    for (d=0; d<n; d++) {
      fprintf(fp," %6.1f",x[d]);
    }
    fprintf(fp,"\n");
  }
  fflush(fp);
}

/************************************************************
	PROC:	lin_eq_solv(A,b,x,n)
	DOES:	Solves a set of n linear equations
************************************************************/

lin_eq_solv(MAT a, float *b, float *x, int n)
{
	int indx[MAX_EQNS];
	float d;
	int n1;
	int flag;

	flag = ludcmp(a,n,indx,&d);
	if (!flag) return(0);
	lubksb(a,n,indx,b);
	for (n1=0; n1<n; n1++)
		x[n1] = b[n1];
	return(1);
}

/************************************************************
	PROC:	ludcmp()
	DOES:	replaces a matrix by the LU decomposition of itself
************************************************************/

#define TINY 1.0e-20

ludcmp(MAT a, int n, int *indx, float *d)
{
	register int i,j,k;
	register float sum;
	int imax;
	float big,dum,temp;
	float *vv;

	vv = vector(1,n);
	*d = 1.0;
	for (i=0; i<n; i++) {
		big = 0.0;
		for (j=0; j<n; j++)
			if ((temp=fabs(a[i][j]))  > big) big=temp;
		if (big==0.0) {
			nrerror("Singular matrix in routine LUDCMP");
			return(0);
		}
		vv[i] = 1.0/big;
	}
	for (j=0; j<n; j++) {
		for (i=0; i<j; i++) {
			sum = a[i][j];
			for (k=0; k<i; k++) sum -= a[i][k]*a[k][j];
			a[i][j] = sum;
		}
		big = 0.0;
		for (i=j; i<n; i++) {
			sum= a[i][j];
			for (k=0; k<j; k++)
				sum -= a[i][k]*a[k][j];
			a[i][j] = sum;
			if ((dum=vv[i]*fabs(sum))>=big) {
				big = dum;
				imax = i;
			}
		}
		if (j!=imax) {
			for (k=0; k<n; k++) {
				dum = a[imax][k];
				a[imax][k] = a[j][k];
				a[j][k] = dum;
			}
			*d = -(*d);
			vv[imax] = vv[j];
		}
		indx[j] = imax;
		if (a[j][j]==0.0) a[j][j]=TINY;
		if (j!=n-1) {
			dum = 1.0/(a[j][j]);
			for (i=j+1; i<n; i++) a[i][j] *= dum;
		}
	}
	free_vector(vv,1,n);
	return(1);
}

/************************************************************
	PROC:	lubksb()
	DOES:	does back-substitution for LU decomp
************************************************************/

lubksb(MAT a, int n, int *indx, float b[])
{
	int i,ii=-1,ip,j;
	float sum;

	for (i=0; i<n; i++) {
		ip=indx[i];
		sum=b[ip];
		b[ip]=b[i];
		if (ii!=-1)
			for (j=ii; j<=i-1; j++) sum -= a[i][j]*b[j];
		else if (sum) ii=i;
		b[i] = sum;
	}
	for (i=n-1; i>=0; i--) {
		sum=b[i];
		for (j=i+1; j<n; j++) sum -= a[i][j]*b[j];
		b[i] = sum/a[i][i];
	}
}

/************************************************************
	PROC:	vector()
	DOES:	allocates a vector
************************************************************/

float *vector(int nl, int nh)
{
	return((float *) malloc(nh*sizeof(float)));
}

/************************************************************
	PROC:	free_vector()
	DOES:	frees the space for a vector
************************************************************/

free_vector(float *v, int nl, int nh)
{
	free((char *) v);
}

/************************************************************
	PROC:	nrerror(str)
	DOES:	numerical recipe error
************************************************************/

nrerror(STR str)
{
	fprintf(stderr,"nrerror: %s\n",str);
	exit(-1);
}

/*********************************************************
    PROC:  matinv(a,size)
    DOES:  inverts a matrix A
**********************************************************/

void matinv(MAT a, int size)
{
  MAT ai,aa,aorig;
  int r,c,c1;
  float mf,pivot;

  matcopy(aorig,a,size);
  for (r=0; r<size; r++) {         /* start with identity inverse matrix */
    for (c=0; c<size; c++) {
      ai[r][c] = (r==c);
    }
  }

  for (r=0; r<size; r++) {   /* for each r */
    for (c=0; c<r; c++) {        /* zero matrix entries before pivot */
      mf = a[r][c];
      for (c1=0; c1<size; c1++) {
	a[r][c1] -= mf*a[c][c1];
	ai[r][c1] -= mf*ai[c][c1]; 
      }
    }
    pivot = a[r][r];              /* get pivot */
    if (pivot==0) {
      printf("matinv: bad pivot\n");
      return;
    }
    mf = 1.0/pivot;
    for (c=0; c<size; c++) {       /* divide r by pivot */
      a[r][c] *= mf;
      ai[r][c] *= mf;
    }
  }
  for (r=0; r<size; r++) {     /* eliminate upper-half */
    for (c=r+1; c<size; c++) {     /* each remaining entry */
      mf = a[r][c];
      for (c1=0; c1<size; c1++) {
	a[r][c1] -= mf*a[c][c1];
	ai[r][c1] -= mf*ai[c][c1] ;
      }
    }
  }
/*
  prmat(a,size);
  prmat(ai,size);
  matmult(aa,aorig,ai,size);
  prmat(aa,size);
*/
  for (r=0; r<size; r++)
    for (c=0; c<size; c++)
      a[r][c] = ai[r][c];
}

/*********************************************************
    PROC:  tm_matinv(a,size)
    DOES:  inverts a matrix A
**********************************************************/

void tm_matinv(Matrix a)
{
  int size=4;
  Matrix ai,aa,aorig;
  int r,c,c1;
  float mf,pivot;

  tm_copy(aorig,a);
  for (r=0; r<size; r++) {         /* start with identity inverse matrix */
    for (c=0; c<size; c++) {
      ai[r][c] = (r==c);
    }
  }

  for (r=0; r<size; r++) {   /* for each r */
    for (c=0; c<r; c++) {        /* zero matrix entries before pivot */
      mf = a[r][c];
      for (c1=0; c1<size; c1++) {
	a[r][c1] -= mf*a[c][c1];
	ai[r][c1] -= mf*ai[c][c1]; 
      }
    }
    pivot = a[r][r];              /* get pivot */
    if (pivot==0) {
      printf("matinv: bad pivot\n");
      return;
    }
    mf = 1.0/pivot;
    for (c=0; c<size; c++) {       /* divide r by pivot */
      a[r][c] *= mf;
      ai[r][c] *= mf;
    }
  }
  for (r=0; r<size; r++) {     /* eliminate upper-half */
    for (c=r+1; c<size; c++) {     /* each remaining entry */
      mf = a[r][c];
      for (c1=0; c1<size; c1++) {
	a[r][c1] -= mf*a[c][c1];
	ai[r][c1] -= mf*ai[c][c1] ;
      }
    }
  }
  for (r=0; r<size; r++)
    for (c=0; c<size; c++)
      a[r][c] = ai[r][c];
}

/*********************************************************
    PROC:  prmat(a,size)
    DOES:  prints a matrix
**********************************************************/

prmat(MAT a, int size)
{
  int r,c;

  printf("\n");
  for (r=0; r<size; r++) {
    for (c=0; c<size; c++) {
      printf("%6.3f ",a[r][c]);
    }
    printf("\n");
  }
}

/*********************************************************
    PROC:  mvmult(x,a,b,size)
    DOES:  multiplies a vector by a matrix
**********************************************************/

mvmult(float *x, MAT a, float *b, int size)
{
  float val;
  int r,c;

  for (r=0; r<size; r++) {
    val=0.0;
    for (c=0; c<size; c++) {
      val += a[r][c]*b[c];
    }
    x[r] = val;
  }
}

/*********************************************************
    PROC:  matcopy(b,a)
    DOES:  copies a matrix
**********************************************************/

matcopy(MAT b, MAT a, int size)
{
  int r,c;

  for (r=0; r<size; r++) {
    for (c=0; c<size; c++) {
      b[r][c] = a[r][c];
    }
  }
}

/*********************************************************
    PROC:  tm_copy(new,old)
    DOES:  copies a matrix
**********************************************************/

tm_copy(GLmatrix b, GLmatrix a)
{
  int r,c;

  for (r=0; r<4; r++) {
    for (c=0; c<4; c++) {
      b[r][c] = a[r][c];
    }
  }
}

/*********************************************************
    PROC:  matmult(aa,a1,a2)
    DOES:  multiplies two matrices
**********************************************************/

matmult(MAT aa, MAT a1,MAT a2, int size)
{
  int r,c;
  float val;
  int c1;

  for (r=0; r<size; r++) {
    for (c=0; c<size; c++) {
      val = 0.0;
      for (c1=0; c1<size; c1++) {
	val += a1[r][c1]*a2[c1][c];
      }
      aa[r][c] = val;
    }
  }
}

/*********************************************************
    PROC:  prfvec(FVEC)
    DOES:  prints type FVEC
**********************************************************/

prfvec(FVEC v)
{
  printf("%6.3f %6.3f %6.3f \n",v[0],v[1],v[2]);
}

/************************************************************
	PROC:	vecsub(ans,a,b)
	DOES:	ans = a - b
************************************************************/

vecsub(FVEC ans, FVEC a, FVEC b)
{
    int d;

    for (d=0; d<3; d++) {
	ans[d] = a[d] - b[d];
    }
}

/************************************************************
	PROC:	vecadd(ans,a,b)
	DOES:	ans = a + b
************************************************************/

vecadd(FVEC ans, FVEC a, FVEC b)
{
    int d;

    for (d=0; d<3; d++) {
	ans[d] = a[d] + b[d];
    }
}

/************************************************************
	PROC:	vecsize(v)
	DOES:	determines a vector size
************************************************************/

float vecsize(FVEC v)
{
    int d;
    float s = 0.0;

    for (d=0; d<3; d++) {
	s += v[d]*v[d];
    }
    return(sqrt(s));
}

/************************************************************
	PROC:	idenmat()
	DOES:	creates an identity matrix
************************************************************/

idenmat(Matrix m)
{
    int i,j;

    for (i=0; i<4; i++) {
	for (j=0; j<4; j++) {
	    m[i][j] = i==j;
	}
    }
}

/************************************************************
	PROC:	tmmult()
	DOES:	multiplies a vector by a transformation matrix
************************************************************/

tmmult(FVEC a, Matrix m, FVEC v)
{
    int i,j;
    float s;

    for (i=0; i<4; i++) {
	s = 0.0;
	for (j=0; j<4; j++) 
	    s += m[i][j]*v[j];
	a[i] = s;
    }
}

/************************************************************
	PROC:	tminv()
	DOES:	inverts a transformation matrix
************************************************************/

tminv(Matrix minv, Matrix m)
{
    int i,j;
    float s;

       /* transpose rotation */
    for (i=0; i<3; i++) {
	for (j=0; j<3; j++) {
	    minv[i][j] = m[j][i];
	}
    }
    for (i=0; i<3; i++) {
	s=0.0;
	for (j=0; j<3; j++) {
	    s += m[j][3]*m[j][i];
	}
	minv[i][3] = -s;
     }
     for (j=0; j<3; j++)
	 minv[3][j] = 0.0;
     minv[3][3] = 1.0;
/*
    tmmmult(mout,minv,m);
    printf("minv check:\n");
    prtm(mout);
*/
 }

/************************************************************
 PROC:	prvec()
 DOES:	prints a vector
************************************************************/

prvec(FVEC v)
{
  int d;

  for (d=0; d<3; d++) {
    printf("%6.3f ",v[d]);
  }
  printf("\n");
}

/************************************************************
  PROC:	prvecf()
  DOES:	prints a vector to a file
************************************************************/

prvecf(FILE *fp, FVEC v)
{
  int d;

  fprintf(fp," %f,%f,%f ",v[0],v[1],v[2]);
}

/************************************************************
  PROC:	getcol(v,m,col)
  DOES:	copies a row of matrix m into v
************************************************************/

getcol(FVEC v, Matrix m, int col)
{
  int n;

  for (n=0; n<3; n++) {
    v[n] = m[n][col];
  }
}

/************************************************************
	 PROC:	putcol(v,m,col)
	 DOES:	copies vector into a column of matrix m
************************************************************/

putcol(FVEC v, Matrix m, int col)
{
  int n;

  for (n=0; n<3; n++) {
    m[n][col] = v[n];
  }
}

/************************************************************
  PROC:	xprod(v,a,b)
  DOES:	calculates a crossproduct
************************************************************/

xprod(FVEC v, FVEC a, FVEC b)
{
     v[0] = a[1]*b[2] - b[1]*a[2];
     v[1] = -a[0]*b[2] + b[0]*a[2];
     v[2] = a[0]*b[1] - b[0]*a[1];
}

/************************************************************
  PROC:	xprodnorm(v,a,b)
  DOES:	finds a normalized orthogonal vector
************************************************************/

xprodnorm(FVEC v, FVEC a, FVEC b)
{
  xprod(v,a,b);
  norm(v);
}

/************************************************************
 PROC:	norm(v)
 DOES:	normalizes a vector
************************************************************/

norm(FVEC v)
{
  float s;

  s = 1.0/sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]);
  v[0] *= s;
  v[1] *= s;
  v[2] *= s;
}

/************************************************************
  PROC:	prtm()
  DOES:	prints a transformation matrix
************************************************************/

prtm(Matrix m)
{
  int i,j;

  for (i=0; i<4; i++) {
    for (j=0; j<4; j++) {
	  printf("%6.3f ",m[i][j]);
	}
    printf("\n");
  }
}

/************************************************************
	 PROC:	prtmf()
	 DOES:	prints a transformation matrix to a file
************************************************************/

prtmf(FILE *f, Matrix m)
{
  int i,j;

  for (i=0; i<4; i++) {
    for (j=0; j<4; j++) {
	  fprintf(f,"%6.3f ",m[i][j]);
	}
    fprintf(f,"\n");
  }
}

/************************************************************
  PROC:	rdtmf()
  DOES:	reads a transformation matrix from a file
************************************************************/

rdtmf(FILE *f, Matrix m)
{
  int i,j;

  for (i=0; i<4; i++) {
    for (j=0; j<4; j++) {
	    fscanf(f,"%f ",&(m[i][j]));
    }
  }
}

/************************************************************
  PROC:	veccp(a,b)
  DOES:	copies b into a
************************************************************/

veccp(FVEC a, FVEC b)
{
  int d;

  for (d=0; d<4; d++)
	 a[d] = b[d];
}

/*********************************************************
  PROC: tmmmult(ans,a,b)
  DOES: multiplies two 4x4 matrices together
**********************************************************/

tmmmult(Matrix ans, Matrix a, Matrix b)
{
  int i,j,n;
  float s;

  for (i=0; i<4; i++) {
    for (j=0; j<4; j++) {
      s = 0.0;
      for (n=0; n<4; n++) {
    	s += a[i][n]*b[n][j];
      }
      ans[i][j] = s;
    }
  }
}

/*********************************************************
  PROC: tmcopy(b,a)
  DOES: copies matrix a into b
**********************************************************/

tmcopy(Matrix b, Matrix a)
{
  int i,j;

  for (i=0; i<4; i++) {
    for (j=0; j<4; j++) {
      b[i][j] = a[i][j];
    }
  }
}


/************************************************************
  PROC:	solv_quad(a,b,c,x1,x2)
  DOES:	solves a quadratic
************************************************************/

solv_quad(float a, float b,
		  float c, float *x1, float *x2)
{
  float k;
  float i;

  k = b*b - 4.0*a*c;
  if (k<0.0) {
    printf("solv_quad: no real roots\n");
  }
  k = sqrt(k);
  i = 0.5/a;
  *x1 = i*(-b + k);
  *x2 = i*(-b - k);
}

/************************************************************
  PROC:	mat2rpy(m,rpy)
  DOES:	calculates roll, pitch, yaw given a transformation matrix
************************************************************/

mat2rpy(Matrix mat, FVEC rpy)
{
  rpy[2] = M_PI + atan2(mat[1][0],mat[0][0]);
  rpy[1] = atan2(-mat[2][0],
                  cos(rpy[2])*mat[0][0] + sin(rpy[2])*mat[1][0]);
  rpy[0] = atan2(sin(rpy[2])*mat[0][2] - cos(rpy[2])*mat[1][2],
                  -sin(rpy[2])*mat[0][1] + cos(rpy[2])*mat[1][1]);
}

/*********************************************************
    PROC: rpy2mat(rpy,mat)
    DOES: calculates a transformation matrix given roll, pitch, yaw
**********************************************************/

rpy2mat(FVEC rpy, Matrix mat)
{
  Matrix mat1, mat2, mat3, tmat;
  FVEC rpy2;

  rotmat('x', rpy[0], mat1 );
  rotmat('y', rpy[1], mat2 );
  rotmat('z', rpy[2], mat3 );
  tmmmult(tmat, mat2, mat1);
  tmmmult(mat, mat3, tmat);
}

/*********************************************************
    PROC: rotmat(axis, rad, mat)
    DOES: generates a rotation matrix
**********************************************************/

rotmat(char axis, float theta, Matrix mat)
{
  float cth, sth;

  cth = cos(theta);
  sth = sin(theta);
  idenmat(mat);
  switch(axis) {
  case 'z':
    mat[0][0] = cth;
    mat[0][1] = -sth;
    mat[1][0] = sth;
    mat[1][1] = cth;
    break;
  case 'y':
    mat[0][0] = cth;
    mat[0][2] = sth;
    mat[2][0] = -sth;
    mat[2][2] = cth;
    break;
  case 'x':
    mat[1][1] = cth;
    mat[1][2] = -sth;
    mat[2][1] = sth;
    mat[2][2] = cth;
    break;
  default:
    printf("rotmat(): ERROR\n");
  }
}

/*********************************************************
    PROC: tmtransp(tpmat, mat)
    DOES: transposes a matrix
**********************************************************/

tmtransp(Matrix tpmat, Matrix mat)
{
  int i;
  int j;

  for (i=0; i<4; i++) {
    for (j=0; j<4; j++) {
      tpmat[i][j] = mat[j][i];
    }
  }
}

/*********************************************************
    PROC: vecinit(v)
    DOES: initializes a vector
**********************************************************/

vecinit(FVEC v)
{
  v[0] = v[1] = v[2] = 0.0;
  v[3] = 1.0;
}

/************************************************************
	PROC:	rotkth(mat,k,th)
	DOES:	determines a matrix from an axis k and an angle th
************************************************************/

rotkth(Matrix mat, FVEC k, float th)
{
    float cth, sth, vth;
    float kx, ky, kz;

    cth = cos(th);
    vth = 1-cth;
    sth = sin(th);
    kx = k[0];
    ky = k[1];
    kz = k[2];
    idenmat(mat);
    mat[0][0] = kx*kx*vth + cth;
    mat[0][1] = ky*kx*vth - kz*sth;
    mat[0][2] = kz*kx*vth + ky*sth;
    mat[1][0] = kx*ky*vth + kz*sth;
    mat[1][1] = ky*ky*vth + cth;
    mat[1][2] = kz*ky*vth - kx*sth;
    mat[2][0] = kx*kz*vth - ky*sth;
    mat[2][1] = ky*kz*vth + kx*sth;
    mat[2][2] = kz*kz*vth + cth;
}

/************************************************************
	PROC:	rotkthi(k, th, mat)
	DOES:	finds k and th from a rotation matrix
************************************************************/

void rotkthi(FVEC k, float *th, Matrix mat)
{
    float sth, cth;
    float a1,a2,a3;

    cth = 0.5*(mat[0][0] + mat[1][1] + mat[2][2] - 1.0);
    a1 = mat[2][1] - mat[1][2];
    a2 = mat[0][2] - mat[2][0];
    a3 = mat[1][0] - mat[0][1];
    sth = 0.5*sqrt(a1*a1 + a2*a2 + a3*a3);
    if (fabs(sth)<0.0001) {
      k[0] = 1.0;
      k[1] = 0.0;
      k[2] = 0.0;
      *th = 0.0;
      return;
    }
    *th = atan2(sth, cth);
    k[0] = (mat[2][1] - mat[1][2])/(2.0*sth);
    k[1] = (mat[0][2] - mat[2][0])/(2.0*sth);
    k[2] = (mat[1][0] - mat[0][1])/(2.0*sth);
}

/*********************************************************
    PROC: mat2q(q,mat)
    DOES: converts a rotation matrix to a quaternion
    FROM: Siggraph Course Notes, 1992, vol. 19, page H22
**********************************************************/

#define MINISCULE 0.0001

mat2q(FVEC q, Matrix m)
{
  float w,x,y,z;
  float r4;

  w = (1 + m[0][0] + m[1][1] + m[2][2])/4.0;
  if (w>MINISCULE) {
    w = sqrt(w);
    r4 = 4.0*w;
    x = (m[2][1] - m[1][2])/r4;
    y = (m[0][2] - m[2][0])/r4;
    z = (m[1][0] - m[0][1])/r4;
  } else {
    w = 0;
    x = -(m[1][1] + m[2][2])/2.0;
    if (x>MINISCULE) {
      x = sqrt(x);
      y = m[1][0]/(2.0*x);
      z = m[2][0]/(2.0*x);
    } else {
      x = 0.0;
      y = (1.0 - m[2][2])/2.0;
      if (y>MINISCULE) {
	y = sqrt(y);
	z = m[2][1]/(2.0*y);
      } else {
	y = 0;
	z = 1;
      }
    }
  }
  q[0] = w;
  q[1] = x;
  q[2] = y;
  q[3] = z;
}

/************************************************************
	PROC:	rot_mat2()
	DOES:	Multiplies the current CTM by the given quaternion rotation
			This is calculated using the minimum number of multiplies
			and additions. Unit quaternions are assumed.
	FROM:	ACM SIGGRAPH '89 Course Notes, "Math for SIGGRAPH", #23,
			page 203
************************************************************/

rot_mat2(float q[], Matrix m)
{
	float w,x,y,z;
	float xs,ys,zs,wx,wy,wz,xx,yy,zz,xy,yz,xz;
	float s;
	
	w = q[0];
	x = q[1];
	y = q[2];
	z = q[3];

/*
	if (w <= -0.999) {
		iden(m);
		return;
	}
*/
	xs = x + x;
	ys = y + y;
	zs = z + z;
	wx = w*xs;
	wy = w*ys;
	wz = w*zs;
	xx = x*xs;
	yy = y*ys;
	zz = z*zs;
	xy = x*ys;
	yz = y*zs;
	xz = x*zs;
	m[0][0] = 1.0 - (yy+zz);
	m[0][1] = xy-wz;
	m[0][2] = xz+wy;
	m[0][3] = 0.0;
	m[1][0] = xy + wz;
	m[1][1] = 1.0 - (xx+zz);
	m[1][2] = yz - wx;
	m[1][3] = 0.0;
	m[2][0] = xz-wy;
	m[2][1] = yz + wx;
	m[2][2] = 1.0 - (xx+yy);
	m[2][3] = 0.0;
	m[3][0] = 0.0;
	m[3][1] = 0.0;
	m[3][2] = 0.0;
	m[3][3] = 1.0;
}
